{-# LANGUAGE TemplateHaskell, RankNTypes, ScopedTypeVariables #-}

module PSO where

import Control.Monad (foldM, replicateM, zipWithM)
import Lens.Simple
import System.Random (Random, randomRIO)

instance Num a => Num [a] where
  (+) = zipWith (+)
  (-) = zipWith (-)
  (*) = zipWith (*)
  abs = map abs
  signum = undefined
  fromInteger x = [fromInteger x]

data Point a b = Point { _position :: a, _score :: b }
  deriving (Eq, Show)
makeLenses ''Point

instance (Eq a, Ord b) => Ord (Point a b) where
  (Point _ s1) `compare` (Point _ s2) = s1 `compare` s2

data Particle a b = Particle
  { _currPoint :: Point a b
  , _bestPoint :: Point a b
  , _velocity :: a
  }
makeLenses ''Particle

data Parameters a = Parameters
  { _omega :: a
  , _phi_p :: a
  , _phi_g :: a
  , _numIterations :: Int
  , _swarmSize :: Int
  }
makeLenses ''Parameters

defaultParameters :: Parameters Double
defaultParameters = Parameters
  { _omega = 0.4
  , _phi_p = 0.4
  , _phi_g = 3.6
  , _numIterations = 10000
  , _swarmSize = 300
  }

type Swarm a b = [Particle a b]


randomVector :: Random r => [(r, r)] -> IO [r]
randomVector = mapM randomRIO

mulVec x = map (*x)

random01Vec dim = randomVector (replicate dim (0, 1))

updateVelocity params best p = do
  rp <- random01Vec dim
  rg <- random01Vec dim
  return $
      (params^.omega) `mulVec` (p^.velocity)
    + (params^.phi_p) `mulVec` (rp * (p^.bestPoint.position - p^.currPoint.position))
    + (params^.phi_g) `mulVec` (rg * (best^.position - p^.currPoint.position))

  where dim = length $ p^.velocity

mkParticle p v = Particle p p v

velocityBounds :: Num a => [(a, a)] -> [(a, a)]
velocityBounds = map (\ x -> (-x, x)) . abs . map (uncurry (-))

initParticle bounds f = do
  pos <- randomVector bounds
  mkParticle <$> pointOfPosition pos f <*> randomVector (velocityBounds bounds)

initSwarm n bounds = replicateM n . initParticle bounds

pointOfPosition pos f = f pos >>= return . Point pos

moveParticle p f v = do
  let pos = (p^.(currPoint.position)) + v
  point <- pointOfPosition pos f
  let best = max (p^.bestPoint) point
  return $ currPoint .~ point $ bestPoint .~ best $ velocity .~ v $ p

maxSwarmPoint :: (Eq a, Ord b) => Swarm a b -> Point a b
maxSwarmPoint = maximum . toListOf (traverse . bestPoint)

addToSwarm params f (swarm, best) p = do
  v <- updateVelocity params best p
  p' <- moveParticle p f v
  return (p' : swarm, max best (p^.bestPoint))

updateSwarm params f (swarm, best) = do
  (s, b) <- foldM (addToSwarm params f) ([], best) swarm
  print b
  return (s, b)

iterateM :: Monad m => Int -> (a -> m a) -> a -> m a
iterateM 0 _ a = return a
iterateM n f a = f a >>= iterateM (n-1) f

runPSO params bounds f = do
  ini <- initSwarm (params^.swarmSize) bounds f
  iterateM (params^.numIterations) (updateSwarm params f) (ini, maxSwarmPoint ini)

griewank :: [Double] -> Double
griewank x = 1
           + 1/4000 * sum (map (^2) x)
           - product (zipWith (\ xi i -> cos $ xi / i) x $ [1 ..])

griebounds = replicate 30 (-600.0, 600.0)

main = runPSO defaultParameters griebounds (return . griewank)
